/*
 * Copyright (c) 2010, Boris Popov <popov.b@gmail.com>
 *
 * Permission to use, copy, modify, and/or distribute
 * this software for any purpose with or without fee
 * is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH
 * THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "easysax.h"
#include "easysaxatt.h"
#include "cstringakk.h"
#include "demirror.h"
#include "rmvdelimiter.h"

static Bool _Fetch (EasyParser);
static char _GetChar (EasyParser);
static void _PassWhiteSpace (EasyParser);
static Bool _HasItEof (EasyParser);
static Bool _IsItNormalNodeBegin (EasyParser);

static Bool _IsItRightBracket (EasyParser);
static Bool _IsItSlashRight (EasyParser);
static Bool _IsItLeftBracket (EasyParser);
static Bool _IsItSpace (EasyParser);
static Bool _IsItQMark (EasyParser);
static Bool _IsItExMark (EasyParser);

static Bool _IsItX (EasyParser);
static Bool _IsItM (EasyParser);
static Bool _IsItL (EasyParser);

static Bool _IsItMinus (EasyParser);

static Bool _IsItE (EasyParser);
static Bool _IsItN (EasyParser);
static Bool _IsItA (EasyParser);

static Bool _IsItLeftFlatBracket (EasyParser);
static Bool _IsItRightFlatBracket (EasyParser);
static Bool _IsItC (EasyParser);

static Bool _IsItD (EasyParser);

static Bool _SearchEndOfNode (EasyParser, StringAkk, StringAkk, Bool*);
static Bool _ParseAttributes (EasyParser, StringAkk);

static Bool _IsIt (EasyParser, const char);

static void _TextOut (EasyParser, StringAkk);
static void _OpenTagOut (EasyParser, StringAkk);
static void _CloseTagOut (EasyParser, StringAkk);
static void _DeclarOut (EasyParser);
static void _ErrorOut (EasyParser, CString);

static CString _Demirror (EasyParser, CString);
static CString _RemoveDelimiter (EasyParser, CString);


EasyParser InitEasyParser (CBuffer cbuffer,
				void (*text) (Text),
				void (*opentag) (OpenTag),
				void (*closetag) (CloseTag),
				void (*attribute) (AttValue),
				void (*declar) (void),
				void (*error) (Error)) {

	EasyParser parser;
	parser = malloc (sizeof (struct _EasyParser));

	if (parser != NULL) {

		parser->Demirror = (void*) InitDeMirror ();		
		parser->RMDelimiter = (void*) InitRemoverDelimiter ();

		if ( ((parser->Demirror) != NULL) &&
				(parser->RMDelimiter != NULL) ) {
			
			parser->buffer = cbuffer;
			parser->FText = text;
			parser->FOpenTag = opentag;
			parser->FCloseTag = closetag;
			parser->FAttribute = attribute;
			parser->FError = error;
			parser->FDeclar = declar;
			parser->First = SUCCESS;

		} else if ( ((parser->Demirror) == NULL) &&
				(parser->RMDelimiter != NULL) ) {

			DestroyRemoverDelimiter
				((RemoverDelimiter) (parser->RMDelimiter));
			free (parser);
			return NULL;

		} else if (((parser->Demirror) != NULL) &&
				(parser->RMDelimiter == NULL)) {

			DestroyDeMirror
				((DeMirror) (parser->Demirror));
			free (parser);
			return NULL;

		} else {

			free (parser);
			return NULL;
		}
	}
	return parser;
}

void DestroyEasyParser (EasyParser parser) {

	if (parser != NULL) {
		
		DestroyDeMirror	((DeMirror) (parser->Demirror));
		DestroyRemoverDelimiter
			((RemoverDelimiter) (parser->RMDelimiter));
		free (parser);
		parser = NULL;
	}
	return;
}

Bool Parse (EasyParser parser) {


begin:
	{
		_PassWhiteSpace (parser);
		if (_HasItEof (parser)) {
			
			return SUCCESS;
		}
		
		if (_IsItLeftBracket (parser)) {
			
			parser->First = FAULT;
			goto notext;
		} 
	
		if (parser->First) {

			_ErrorOut (parser, "Incorrect beginning of the document\n");
			return FAULT;
		}
		goto text;	
	}


text:
	{
		StringAkk sa = InitStringAkk ();
		AddByteToStringAkk (sa, _GetChar (parser));
		while (_Fetch (parser)) {
			
			if (_IsItLeftBracket (parser)) {
				
				_TextOut (parser, sa);
				DestoryStringAkk (sa);
				goto notext;
			}
			
			AddByteToStringAkk (sa, _GetChar (parser));
		}
		_ErrorOut (parser, "Incorrect ending of the document\n");
		DestoryStringAkk (sa);
		return FAULT;
	}


notext:
	{
		if (_Fetch (parser)) {
			
			if (_IsItNormalNodeBegin (parser)) {
				
				goto beginnode;
			
			} else if (_IsItSlashRight (parser)) {
				
				goto closenode;
			
			} else {
				
				goto nobeginnode;
			}
		}
		_ErrorOut (parser, "Incorrect ending of the tag\n");
		return FAULT;
	}


closenode:
	{
		StringAkk sa = InitStringAkk ();
		while (_Fetch (parser)) {
			
			if (_IsItRightBracket (parser)) {
				
				_CloseTagOut (parser, sa);
				DestoryStringAkk (sa);
				goto begin;
			}
			
			AddByteToStringAkk (sa, _GetChar (parser));
		}
		_ErrorOut (parser, "Incorrect ending of the close tag\n");
		DestoryStringAkk (sa);
		return FAULT;
	}


beginnode:
	  {
		StringAkk sa = InitStringAkk ();
		AddByteToStringAkk (sa, _GetChar (parser));
		while (_Fetch (parser)) {
			
			if (_IsItSlashRight (parser)) {

				if (_Fetch (parser)) {

					if (_IsItRightBracket (parser)) {

						_OpenTagOut (parser, sa);
						_CloseTagOut (parser, sa);
						DestoryStringAkk (sa);
						goto begin;
					}
					break;
				}
				break;

			} else if (_IsItRightBracket (parser)) {

				_OpenTagOut (parser, sa);
				DestoryStringAkk (sa);
				goto begin;
			}

			else if (_IsItSpace (parser)) {

				_PassWhiteSpace (parser);
				if (!_HasItEof (parser)) {
			
					if (_IsItSlashRight (parser)) {
						
						if (_Fetch (parser)) {
							
							if (_IsItRightBracket (parser)) {

								//нет атрибутов!
								_OpenTagOut (parser, sa);
								_CloseTagOut (parser, sa);
								DestoryStringAkk (sa);
								goto begin;
							}
							break;
						}
						break;
					}

					if (_IsItRightBracket (parser)) {
						
						_OpenTagOut (parser, sa);
						DestoryStringAkk (sa);
						goto begin;
					}

					//возможно наличие атрибутов
					StringAkk attrib = InitStringAkk ();
					Bool PRINTCLOSE = FAULT;
					if (_SearchEndOfNode (parser, attrib, sa, &PRINTCLOSE)) {
						
						if (_ParseAttributes (parser, attrib)) {

							if (PRINTCLOSE) _CloseTagOut (parser, sa);
							DestoryStringAkk (sa);
							DestoryStringAkk (attrib);
							goto begin;
						}
						_ErrorOut (parser, "Error of parsing attributes\n");
						DestoryStringAkk (attrib);
						break;
					}
					DestoryStringAkk (attrib);
					break;
				}
				break;
			}
			AddByteToStringAkk (sa, _GetChar (parser));
		}
		_ErrorOut (parser, "Incorrect ending of the document\n");
		DestoryStringAkk (sa);
		return FAULT;
	  }


nobeginnode:
	  {
		  if (_IsItQMark (parser)) {

			  goto pi_or_de;
		  }

		  if (_IsItExMark (parser)) {

			  goto other;
		  }
		  _ErrorOut (parser, "Unrecognized statement\n");
		  return FAULT;
	  }


pi_or_de:
	  {
		  if (_Fetch (parser)) {

			  if (_IsItX (parser)) {
				  
				  if (_Fetch (parser)) {
					  
					  if (_IsItM (parser)) {
						  
						  if (_Fetch (parser)) {

							  if (_IsItL (parser)) {

								  goto de;
							  }
							  goto pi;
						  }
						  _ErrorOut (parser, "Incorrect ending of the document\n");
						  return FAULT;
					  }
					  goto pi;
				  }
				  _ErrorOut (parser, "Incorrect ending of the document\n");
				  return FAULT;
			  }
			  goto pi;
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


other:
	  {
		  if (_Fetch (parser)) {

			  if (_IsItMinus (parser)) {
				  
				  if (_Fetch (parser)) {
					  
					  if (_IsItMinus (parser)) {

						  goto comment;
					  }
					  _ErrorOut (parser, "Unrecognized statement\n");
					  return FAULT;
				  }
				  _ErrorOut (parser, "Incorrect ending of the document\n");
				  return FAULT;
			  }

			  if (_IsItE (parser)) goto element;
			  if (_IsItN (parser)) goto notation;
			  if (_IsItA (parser)) goto attlist;

			  if (_IsItLeftFlatBracket (parser)) {
		  
				  if (_Fetch (parser)) {

					  if (_IsItC (parser)) goto cdata;
					  goto ifblock;
				  }
				  _ErrorOut (parser, "Incorrect ending of the document\n");
				  return FAULT;
			  }

			  if (_IsItD (parser)) goto doctype;
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


notation:
attlist:
element:
	  {
		  /* TODO compare ENTITY vs ELEMENT */
		  while (_Fetch (parser)) {
			  
			  if (_IsItRightBracket (parser)) {

				  goto begin;
			  }
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


cdata:
	  {
		  while (_Fetch (parser)) {
			  
			  if (_IsItRightFlatBracket (parser)) {
				  
				  if (_Fetch (parser)) {
					  
					  if (_IsItRightFlatBracket (parser)) {
						  
						  if (_Fetch (parser)) {
							  
							  if (_IsItRightBracket (parser)) {

								  goto begin;
							  }
							  continue;
						  }
						  break;
					  }
					  continue;
				  }
				  break;
			  }
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


ifblock:
	  {
		  goto cdata;
	  }


doctype:
	  {
		  size_t count_left_bracket = 1;
		  while (_Fetch (parser)) {

			  if (_IsItLeftBracket (parser)) {
				  
				  count_left_bracket++;
			  }
			  
			  if (_IsItRightBracket (parser)) {
				  
				  count_left_bracket--;
			  }

			  if (count_left_bracket == 0) {

				  goto begin;
			  }
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


comment:
	  {
		  while (_Fetch (parser)) {

			  if (_IsItMinus (parser)) {
		  
				  if (_Fetch (parser)) {
			  
					  if (_IsItMinus (parser)) {
						  
						  if (_Fetch (parser)) {
							  
							  if (_IsItRightBracket (parser)) {

								  goto begin;
							  }
							  continue;
						  }
						  break;
					  }
					  continue;
				  }
				  break;
			  }
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


pi:
	  {
		  while (_Fetch (parser)) {

			  if (_IsItQMark (parser)) {
		  
				  if (_Fetch (parser)) {

					  if (_IsItRightBracket (parser)) {

						  goto begin;
					  }
					  break;
				  }
				  break;
			  }
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  return FAULT;
	  }


de:
	  {
		  StringAkk attrib = InitStringAkk ();
		  if (attrib == NULL) {

			  return FAULT;
		  }

		  while (_Fetch (parser)) {

			  if (_IsItQMark (parser)) {
		  
				  if (_Fetch (parser)) {

					  if (_IsItRightBracket (parser)) {

						  _DeclarOut (parser);
						  if (_ParseAttributes (parser, attrib)) {
							  
							  DestoryStringAkk (attrib);
							  goto begin;
						  }
						  break;
					  }
					  break;
				  }
				  break;
			  }
			  AddByteToStringAkk (attrib, _GetChar (parser));
		  }
		  _ErrorOut (parser, "Incorrect ending of the document\n");
		  DestoryStringAkk (attrib);
		  return FAULT;
	  }
}

static Bool _ParseAttributes (EasyParser parser, StringAkk akk) {

	CBuffer buffer =
		InitCBufferFromString (GetStringFromStringAkk (akk));
	if (buffer == NULL) {

		return FAULT;
	}

	EasyParserAtt parserAtt =
		InitEasyParserAtt (buffer, parser->FAttribute,
				parser->Demirror, parser->RMDelimiter);
	if (parserAtt == NULL) {

		DestroyCBuffer (buffer);
		return FAULT;
	}

	Bool RESULT = FAULT;

	if (ParseAtt (parserAtt)) {

		RESULT = SUCCESS;
	}

	DestroyEasyParserAtt (parserAtt);
	DestroyCBuffer (buffer);
	return RESULT;
}

static Bool _SearchEndOfNode
(EasyParser parser, StringAkk akk, StringAkk node, Bool* PRINTCLOSE) {

	AddByteToStringAkk (akk, _GetChar (parser));
	
	while (_Fetch (parser)) {
							
		if (_IsItRightBracket (parser)) {

			_OpenTagOut (parser, node);
			return SUCCESS;
		}
					
		if (_IsItSlashRight (parser)) {
						
			if (_Fetch (parser)) {
			
				if (_IsItRightBracket (parser)) {

					_OpenTagOut (parser, node);
					*(PRINTCLOSE) = SUCCESS;
					return SUCCESS;
				}
				continue;
			}
			break;
		}
		AddByteToStringAkk (akk, _GetChar (parser));
	}
	return FAULT;
}

static CString _Demirror (EasyParser parser, CString string) {

	DeMirror demirror = (DeMirror) parser->Demirror;
	return (DecodeDeMirror (demirror, string));
}

static CString _RemoveDelimiter
(EasyParser parser, CString string) {

	RemoverDelimiter rd = (RemoverDelimiter) parser->RMDelimiter;
	return (RemoveDelimiter (rd, string));
}

static void _TextOut (EasyParser parser, StringAkk akk) {

	Out (parser->FText,
			_Demirror (parser,
				_RemoveDelimiter (parser,
					GetStringFromStringAkk (akk))));
	return;
}

static void _OpenTagOut (EasyParser parser, StringAkk akk) {

	Out (parser->FOpenTag,
			_RemoveDelimiter (parser,
				GetStringFromStringAkk (akk)));
	return;
}

static void _CloseTagOut (EasyParser parser, StringAkk akk) {

	Out (parser->FCloseTag,
			_RemoveDelimiter (parser,
				GetStringFromStringAkk (akk)));
	return;
}

static void _ErrorOut (EasyParser parser, CString string) {

	Out (parser->FError, string);
	return;
}

static Bool _Fetch (EasyParser parser) {
	
	parser->result = GetChar (parser->buffer);
	if (_HasItEof (parser)) {

		return FAULT;
	}
	return SUCCESS;
}

static Bool _HasItEof (EasyParser parser) {
	
	return HasItEof (parser->result);
}

static char _GetChar (EasyParser parser) {
	
	return GetCharFromResult (parser->result);
}

static void _PassWhiteSpace (EasyParser parser) {
	
	while (_Fetch (parser)) {
		
		if (! _IsItSpace (parser)) {

			break;
		}
	}
	return;
}

static Bool _IsItNormalNodeBegin (EasyParser parser) {
	
	if ( ! ((_IsItExMark     (parser)) ||
			(_IsItQMark      (parser)) ||
			(_IsItSlashRight (parser)))) {

		return SUCCESS;
	}
	return FAULT;
}

static Bool _IsItLeftBracket (EasyParser parser) {

	return _IsIt (parser, '<');
}

static Bool _IsItRightBracket (EasyParser parser) {
	
	return _IsIt (parser, '>');
}

static Bool _IsItSlashRight (EasyParser parser) {

	return _IsIt (parser, '/');
}

static Bool _IsIt (EasyParser parser, const char ch) {

	return IsIt (_GetChar (parser), ch);
}

static Bool _IsItSpace (EasyParser parser) {
	
	char ch2 = _GetChar (parser);
	return IsItWhiteSpace (ch2);
}

static Bool _IsItQMark (EasyParser parser) {

	return _IsIt (parser, '?');
}

static Bool _IsItExMark (EasyParser parser) {
	
	return _IsIt (parser, '!');
}

static Bool _IsItX (EasyParser parser) {
	
	return ((_IsIt (parser, 'x')) || (_IsIt (parser, 'X')));
}

static Bool _IsItM (EasyParser parser) {
	
	return ((_IsIt (parser, 'm')) || (_IsIt (parser, 'M')));
}

static Bool _IsItL (EasyParser parser) {
	
	return ((_IsIt (parser, 'l')) || (_IsIt (parser, 'L')));
}

static Bool _IsItMinus (EasyParser parser) {
	
	return _IsIt (parser, '-');
}

static Bool _IsItE (EasyParser parser) {
	
	return _IsIt (parser, 'E');
}

static Bool _IsItN (EasyParser parser) {
	
	return _IsIt (parser, 'N');
}

static Bool _IsItA (EasyParser parser) {
	
	return _IsIt (parser, 'A');
}

static Bool _IsItLeftFlatBracket (EasyParser parser) {
	
	return _IsIt (parser, '[');
}

static Bool _IsItRightFlatBracket (EasyParser parser) {
	
	return _IsIt (parser, ']');
}

static Bool _IsItC (EasyParser parser) {
	
	return _IsIt (parser, 'C');
}

static Bool _IsItD (EasyParser parser) {
	
	return _IsIt (parser, 'D');
}

static void _DeclarOut (EasyParser parser) {

	if ((parser->FDeclar) != NULL) {
	
		parser->FDeclar ();
	}
	return;
}
